using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using RiffParserDemo2;
using System.ComponentModel;

namespace Nash.Multimedia.Riff
{
    class ElementRemover
    {
        private String inFile;
        private String outFile;
        private String tempFile;
        private FileStream fileStream;
        private BinaryWriter binaryWriter;
        private int bytesSkipped = 0;
        private String[] unwantedLists;
        private String[] unwantedChunks;
        private RiffParser riffParser;       

        // Constructors for ElementRemover
        public ElementRemover(String inFile, String outFile)
        {
            setup(inFile, outFile);
        }

        public ElementRemover(String inFile)
        {
            setup(inFile, inFile);
        }

        // Establish the filenames used.
        private void setup(String inFile, String outFile)
        {
            this.inFile = inFile;
            this.outFile = outFile;

            // Ensure the temp file is not named the same
            // as any other file already existing.
            int i = 0;
            while (i >= 0)
            {
                tempFile = "temp#" + i + ".tmp";
                if (File.Exists(tempFile))
                {
                    i++;
                }
                else break;
            }
        }

        /// <summary>
        /// Close all streams and delete all temporary files.
        /// </summary>
        public void cleanUp()
        {
            // Close the streams.
            riffParser.CloseFile();
            binaryWriter.Close();
            fileStream.Close();

            // Delete the tempfile.
            File.Delete(tempFile);
        }

        /// <summary>
        /// Read through the RIFF file and copy its contents, excluding the unwanted Lists.
        /// Then adjust the FileSize value within the header to reflect the changes.
        /// </summary>
        /// <param name="targetLists">FourCCs of lists to remove.</param>
        public void removeLists(String[] targetLists)
        {
            // We are not removing any chunks, so there are zero unwanted chunks.
            String[] chunks = new String[0];
            this.removeElements(targetLists, chunks);
        }

        /// <summary>
        /// Read through the RIFF file and copy its contents, excluding the unwanted Chunks.
        /// Then adjust the FileSize value within the header to reflect the changes.
        /// </summary>
        /// <param name="targetChunks">FourCCs of chunks to remove.</param>
        public void removeChunks(String[] targetChunks)
        {
            // We are not removing any lists, so there are zero unwanted lists.
            String[] lists = new String[0];
            this.removeElements(lists, targetChunks);
        }

        /// <summary>
        /// Read through the RIFF file and copy its contents, excluding the unwanted Elements.
        /// Then adjust the FileSize value within the header to reflect the changes.
        /// </summary>
        /// <param name="targetLists">FourCCs of lists to remove.</param>
        /// <param name="targetChunks">FourCCs of chunks to remove.</param>
        public void removeElements(String[] targetLists, String[] targetChunks)
        {
            // Define the Lists and Chunks to be removed.
            unwantedLists = targetLists;
            unwantedChunks = targetChunks;

            // Create a new FileStream to write to.
            fileStream = new FileStream(tempFile, FileMode.Append, FileAccess.Write);
            
            // Create a new Binary Writing Stream.
            binaryWriter = new BinaryWriter(fileStream);

            // Create a new RiffParser instance
            riffParser = new RiffParser();

            // Open the file
            riffParser.OpenFile(inFile);

            // If this point is reached, the file is valid.

            // Store the size to loop through elements.
            int size = riffParser.DataSize;

            // Define the processing delegates.
            RiffParser.ProcessChunkElement processChunk = new RiffParser.ProcessChunkElement(ProcessChunk);
            RiffParser.ProcessListElement processList = new RiffParser.ProcessListElement(ProcessList);

            // Write RIFF header.
            binaryWriter.Write(riffParser.FileRIFF);
            binaryWriter.Write(riffParser.DataSize);
            binaryWriter.Write(riffParser.FileType);

            // Read all top level elements and chunks.
            try
            {
                while (size > 0)
                {
                    // Get the next element if there is one.
                    if (false == riffParser.ReadElement(ref size, processChunk, processList))
                    {
                        break;
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("Problem: " + ex.ToString());
            }

            // Set stream postion to the FileSize property in the header.
            fileStream.Seek(RiffParser.DWORDSIZE, SeekOrigin.Begin);

            // Overwrite the FileSize with the updated value.
            binaryWriter.Write(riffParser.DataSize - bytesSkipped);

            // Close Streams
            binaryWriter.Close();
            fileStream.Close();

            // Copy from the temp file to the target output.
            File.Copy(tempFile, outFile, true);

            // Cleanup everything before finishing
            riffParser.CloseFile();
            File.Delete(tempFile);
        }

        // Process a RIFF List element.
        private void ProcessList(RiffParser riffParser, int fourCC, int length)
        {           
            if (unwantedLists.Length == 0)
            {
                // There are no unwanted lists, so we always copy them.
                CopyList(riffParser, fourCC, length);
            }
            else
            {
                for (int i = 0; i < unwantedLists.Length; i++)
                {
                    if (0 == String.Compare(RiffParser.FromFourCC(fourCC), unwantedLists[i]))
                    {
                        // This is an unwanted List. Ignore it and keep track
                        // of the number of bytes removed.
                        riffParser.SkipData(length);
                        bytesSkipped += (length + RiffParser.DWORDSIZE);                   
                    }
                    else
                    {
                        // This is a wanted list so we copy it.
                        CopyList(riffParser, fourCC, length);
                    }
                }
            }
        }

        // Process a RIFF Chunk element.
        private void ProcessChunk(RiffParser riffParser, int fourCC, int length, int paddedLength)
        {
            if (unwantedChunks.Length == 0)
            {
                // There are no unwanted chunks, so we always copy them.
                CopyChunk(riffParser, fourCC, length, paddedLength);
            }
            else
            {
                for (int i = 0; i < unwantedChunks.Length; i++)
                {
                    if (0 == String.Compare(RiffParser.FromFourCC(fourCC), unwantedChunks[i]))
                    {
                        // This is an unwanted Chunk. Ignore it and keep track
                        // of the number of bytes removed.
                        riffParser.SkipData(paddedLength);
                        bytesSkipped += (paddedLength + RiffParser.DWORDSIZE);
                    }
                    else
                    {
                        // This is a wanted chunk so we copy it.
                        CopyChunk(riffParser, fourCC, length, paddedLength);
                    }
                }
            }
        }

        private void CopyList(RiffParser riffParser, int fourCC, int length)
        {
            // Write List to file.
            binaryWriter.Write(RiffParser.ToFourCC(RiffParser.LIST4CC));
            binaryWriter.Write(length + RiffParser.DWORDSIZE);
            binaryWriter.Write(fourCC);

            // Define the processing delegates.
            RiffParser.ProcessChunkElement processChunk = new RiffParser.ProcessChunkElement(ProcessChunk);
            RiffParser.ProcessListElement processList = new RiffParser.ProcessListElement(ProcessList);

            // Read all elements within list.
            try
            {
                while (length > 0)
                {
                    // Get the next element if there is one.
                    if (false == riffParser.ReadElement(ref length, processChunk, processList))
                    {
                        break;
                    }
                }
            }
            catch (Exception ex)
            {
            }
        }

        private void CopyChunk(RiffParser riffParser, int fourCC, int length, int paddedLength)
        {
            // Initialise array used to temporarily store the data read
            // before it is copied into the new file.
            byte[] byteArray = new byte[1];

            // Write Chunk to file.
            binaryWriter.Write(fourCC);
            binaryWriter.Write(length);

            // Read the data and write to file.
            for (int j = 0; j < paddedLength; j++)
            {
                riffParser.ReadData(byteArray, 0, 1);
                binaryWriter.Write(byteArray[0]);
            }
        }
    }
}
